#include "type_system.h"
#include <stdio.h>

struct Option* T_define_option()
{
  struct Option *opt;

  opt = (struct Option*)calloc(1, sizeof(struct Option));
  opt->key = NULL;
  opt->type = NULL;
  opt->description = NULL;
  opt->answer = NULL;

  return opt;
}

/*Base type functions*/
struct T_type* T_type_alloc()
{
  struct T_type *t;
  t = (struct T_type*)calloc(1, sizeof(struct T_type));
  t->p = NULL;
  t->print = NULL;
  
  return t;
}

struct T_type* T_boolean(enum T_BOOL value)
{
  struct T_type *t;
  struct T_type_boolean *bt;

  t =  T_type_alloc();
  
  bt = (struct T_type_boolean*)calloc(1, sizeof(struct T_type_boolean));
  bt->value = value;

  t->type = BOOLEAN;
  t->p = bt;
  t->print = T_boolean_print;

  return t;
}

void T_boolean_print(struct T_type* type)
{
  struct T_type_boolean* p;
  p = (struct T_type_boolean*)type->p;
  if(p->value == True)
    fprintf(stderr, "TRUE\n");
  else
    fprintf(stderr, "FALSE\n");

  return;
}

struct T_type* T_integer(size_t value)
{
  struct T_type *t;
  struct T_type_integer *it;

  t =  T_type_alloc();
  
  it = (struct T_type_integer*)calloc(1, sizeof(struct T_type_integer));
  it->value = value;

  t->type = INTEGER;
  t->p = it;
  t->print = T_integer_print;

  return t;
}

void T_integer_print(struct T_type* type)
{
  struct T_type_integer* p;
  p = (struct T_type_integer*)type->p;
  if(p)
    fprintf(stderr, "%i\n", p->value);

  return;
}

struct T_type* T_integer_subrange(size_t first, size_t last, size_t step)
{
  struct T_type *t;
  struct T_type_integer_subrange *it;

  t =  T_type_alloc();
  
  it = (struct T_type_integer_subrange*)calloc(1, sizeof(struct T_type_integer_subrange));
  it->first = first;
  it->last = last;
  it->step = step;

  t->type = INTEGER_SUBRANGE;
  t->p = it;
  t->print = T_integer_subrange_print;

  return t;
}

void T_integer_subrange_print(struct T_type* type)
{
  struct T_type_integer_subrange* p;
  p = (struct T_type_integer_subrange*)type->p;
  if(p) {
    fprintf(stderr, "first: %i\n", p->first);
    fprintf(stderr, "last : %i\n", p->last);
    fprintf(stderr, "step : %i\n", p->step);
  }

  return;
}


struct T_type* T_integer_enumeration(size_t count, size_t *values)
{
  struct T_type *t;
  struct T_type_integer_enumeration *it;

  t =  T_type_alloc();
  
  it = (struct T_type_integer_enumeration*)calloc(1, sizeof(struct T_type_integer_enumeration));
  it->count = count;
  it->values = values;

  t->type = INTEGER_ENUMERATION;
  t->p = it;
  t->print = T_integer_enumeration_print;

  return t;
}

void T_integer_enumeration_print(struct T_type* type)
{
  int i;
  struct T_type_integer_enumeration* p;
  p = (struct T_type_integer_enumeration*)type->p;
  if(p)
    for(i = 0; i < p->count; i++) {
      fprintf(stderr, "%i\n", p->values[i]);
    }

  return;
}

struct T_type* T_real(double value)
{
  struct T_type *t;
  struct T_type_real *it;

  t =  T_type_alloc();
  
  it = (struct T_type_real*)calloc(1, sizeof(struct T_type_real));
  it->value = value;

  t->type = REAL;
  t->p = it;
  t->print = T_real_print;

  return t;
}

void T_real_print(struct T_type* type)
{
  struct T_type_real* p;
  p = (struct T_type_real*)type->p;
  if(p)
    fprintf(stderr, "%f\n", p->value);

  return;
}


struct T_type* T_real_subrange(double first, enum T_BOOL first_inclusive, double last, enum T_BOOL last_inclusive)
{
  struct T_type *t;
  struct T_type_real_subrange *it;

  t =  T_type_alloc();
  
  it = (struct T_type_real_subrange*)calloc(1, sizeof(struct T_type_real_subrange));
  it->first = first;
  it->last = last;
  it->first_inclusive = first_inclusive;
  it->last_inclusive = last_inclusive;

  t->type = REAL_SUBRANGE;
  t->p = it;
  t->print = T_real_subrange_print;

  return t;
}

void T_real_subrange_print(struct T_type* type)
{
  struct T_type_real_subrange* p;
  p = (struct T_type_real_subrange*)type->p;
  if(p) {
    fprintf(stderr, "first: %f\n", p->first);
    fprintf(stderr, "last: %f\n", p->last);
    fprintf(stderr, "first_include: %i\n", p->first_inclusive);
    fprintf(stderr, "last_include: %i\n", p->last_inclusive);
  }

  return;
}

struct T_type* T_string(const char *value)
{
  struct T_type *t;
  struct T_type_string *it;

  t =  T_type_alloc();
  
  it = (struct T_type_string*)calloc(1, sizeof(struct T_type_string));
  it->value = value;

  t->type = STRING;
  t->p = it;
  t->print = T_string_print;

  return t;
}

void T_string_print(struct T_type* type)
{
  struct T_type_string* p;
  p = (struct T_type_string*)type->p;
  if(p && p->value)
    fprintf(stderr, "%s\n", p->value);

  return;
}


struct T_type* T_string_enumeration(size_t count, struct T_type_string** values)
{
  struct T_type *t;
  struct T_type_string_enumeration *it;

  t =  T_type_alloc();
  
  it = (struct T_type_string_enumeration*)calloc(1, sizeof(struct T_type_string_enumeration));
  it->count = count;
  it->values = values;

  t->type = STRING_ENUMERATION;
  t->p = it;

  return t;

}
/*
struct T_type* T_string_enumeration_with_descs(size_t count, char** values, char **descs);
*/
struct T_type* T_element(enum T_ELEMENT element, enum T_AGE age)
{
  struct T_type *t;
  struct T_type_element *it;

  t =  T_type_alloc();
  
  it = (struct T_type_element*)calloc(1, sizeof(struct T_type_element));
  it->element = element;
  it->age = age;

  t->type = STRING;
  t->p = it;

  return t;
}
/* Composite types functions */
/*
struct T_type* T_optional(struct T_type* base);
struct T_type* T_tuple(size_t count, struct T_type** types);
struct T_type* T_record(size_t count, char** names, struct T_type** types);
struct T_type* T_list(struct T_type* type, enum T_BOOL allow_empty, enum T_BOOL allow_duplicates, enum T_BOOL order_matters);
*/
struct T_type* T_union(size_t count, struct T_type** types)
{
  struct T_type *t;
  struct T_type_tuple *it;

  if(count < 2)
    return NULL;

  t =  T_type_alloc();
  
  it = (struct T_type_tuple*)calloc(1, sizeof(struct T_type_tuple));
  it->count = count;
  it->types = types;

  t->type = UNION;
  t->p = it;

  return t;
}

