#include <stdlib.h>

struct Option {
  struct T_type *key;
  struct T_type *type;
  struct T_type *description;
  struct T_type *answer;
};

/*Type library enumerations*/
enum T_BOOL {
  False,
  True
};

/*What kind of elements exist in grass*/
enum T_ELEMENT {
  RASTER,
  VECTOR,
  FILENAME,
  VOLUME,
  GROUP
};

/*The age of a datatype*/
enum T_AGE {
  NEW,
  OLD
};

/*All available types*/
enum T_TYPES {
  BOOLEAN,
  INTEGER,
  INTEGER_SUBRANGE,
  INTEGER_ENUMERATION,
  REAL,
  REAL_SUBRANGE,
  STRING,
  STRING_ENUMERATION,
  STRING_ENUMERATION_WITH_DESCS,
  ELEMENT,
  TUPLE,
  RECORD,
  LIST,
  UNION
};

/*Base type structures */
struct T_type {
  enum T_TYPES type;
  void *p; /*Pointer to a type structure*/
  void (*print)(struct T_type*);
};

struct T_type_boolean {
  enum T_BOOL value;
};

struct T_type_integer {
  size_t value;
};

struct T_type_integer_subrange {
  size_t first;
  size_t last;
  size_t step;
};

struct T_type_integer_enumeration {
  size_t count;
  size_t *values;
};

struct T_type_real {
  double value;
};

struct T_type_real_subrange {
  double first;
  double last;
  enum T_BOOL first_inclusive;
  enum T_BOOL last_inclusive;
};

struct T_type_string {
  size_t size;
  const char *value;
}; 

struct T_type_string_enumeration {
  size_t count;
  struct T_type_string **values;
}; 

struct T_type_string_enumeration_with_descs {
  int count;
  struct T_type_string **values;
  struct T_type_string **descs;
}; 

struct T_type_element {
  enum T_ELEMENT element;
  enum T_AGE age;
};

/*Complex type structures */
struct T_type_optional {
  struct T_type *base;
};

struct T_type_tuple {
  size_t count;
  struct T_type **types;
};

struct T_type_record {
  size_t count;
  struct T_type **record;
  struct T_string **names;
};

struct T_type_list {
  struct T_type *type;
  enum T_BOOL allow_empty;
  enum T_BOOL allow_duplicates;
  enum T_BOOL order_matters;
};

/*Base type functions*/
struct Option* T_define_option();

struct T_type* T_type_alloc();
struct T_type* T_boolean(enum T_BOOL value);
void T_boolean_print(struct T_type* type);
struct T_type* T_integer(size_t value);
void T_integer_print(struct T_type* type);
struct T_type* T_integer_subrange(size_t first, size_t last, size_t step);
void T_integer_subrange_print(struct T_type* type);
struct T_type* T_integer_enumeration(size_t count, size_t *values);
void T_integer_enumeration_print(struct T_type* type);
struct T_type* T_real(double value);
void T_real_print(struct T_type* type);
struct T_type* T_real_subrange(double first, enum T_BOOL first_inclusive, double last, enum T_BOOL last_inclusive);
void T_real_subrange_print(struct T_type* type);
struct T_type* T_string(const char *value);
void T_string_print(struct T_type* type);
struct T_type* T_string_enumeration(size_t count, struct T_type_string** values);
struct T_type* T_string_enumeration_with_descs(size_t count, struct T_type_string** values, struct T_type_string** descs);
struct T_type* T_element(enum T_ELEMENT element, enum T_AGE age);

/* Composite types functions */
struct T_type* T_optional(struct T_type* base);
struct T_type* T_tuple(size_t count, struct T_type** types);
struct T_type* T_record(size_t count, char** names, struct T_type** types);
struct T_type* T_list(struct T_type* type, enum T_BOOL allow_empty, enum T_BOOL allow_duplicates, enum T_BOOL order_matters);
struct T_type* T_union(size_t count, struct T_type** types);



